<?php
namespace Drupal\commerce_pvt\Plugin\views\field;

use Drupal\commerce_cart\CartManagerInterface;
use Drupal\commerce_cart\CartProviderInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Resolver\ChainOrderTypeResolverInterface;
use Drupal\commerce_order\Resolver\ChainOrderItemTypeResolverInterface;
use Drupal\commerce_price\CurrencyFormatter;
use Drupal\commerce_pvt\RefreshPageElementsHelper;
use Drupal\commerce_store\CurrentStoreInterface;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\views\Plugin\views\field\FieldPluginBase;
use Drupal\views\Plugin\views\field\UncacheableFieldHandlerTrait;
use Drupal\views\ResultRow;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a form element for editing the order item quantity.
 *
 * @ViewsField("commerce_product_variation_edit_quantity")
 */
class EditVariationQuantity extends FieldPluginBase {

  use UncacheableFieldHandlerTrait;

  /**
   * The Order Type Chain resolver.
   *
   * @var \Drupal\commerce_order\Resolver\ChainOrderTypeResolverInterface
   */
  protected $chainOrderTypeResolver;

  /**
   * The Order Item Type Chain resolver.
   *
   * @var \Drupal\commerce_order\Resolver\ChainOrderItemTypeResolverInterface
   */
  protected $chainOrderItemTypeResolver;

  /**
   * The cart manager.
   *
   * @var \Drupal\commerce_cart\CartManagerInterface
   */
  protected $cartManager;

  /**
   * The Cart provider.
   *
   * @var \Drupal\commerce_cart\CartProviderInterface
   */
  protected $cartProvider;

  /**
   * The Current store resolved,
   *
   * @var \Drupal\commerce_store\Entity\StoreInterface
   */
  protected $currentStore;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * RefreshPageElementsHelper service.
   *
   * @var \Drupal\dc_ajax_add_cart\RefreshPageElementsHelper
   */
  protected $refreshPageElementsHelper;

  /**
   * The currency formatter for this class
   *
   * @var \Drupal\commerce_price\CurrencyFormatter
   */
  protected $currencyFormatter;


  /**
   * Constructs a new EditQuantity object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\commerce_order\Resolver\ChainOrderTypeResolverInterface $chain_order_type_resolver
   *   The chain resolver for product order type.
   * @param \Drupal\commerce_order\Resolver\ChainOrderItemTypeResolverInterface $chain_order_item_type_resolver
   *   The chain resolver for product order item type.
   * @param \Drupal\commerce_cart\CartManagerInterface $cart_manager
   *   The cart manager.
   * @param \Drupal\commerce_cart\CartProviderInterface $cart_provider
   *   The cart provider.
   * @param \Drupal\commerce_store\CurrentStoreInterface $current_store
   *   The current store resolver.
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match object.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Drupal\commerce_price\CurrencyFormatter $currency_formatter
   *   The currency formatter.
   */
  public function __construct(array $configuration,
                              $plugin_id,
                              $plugin_definition,
                              ChainOrderTypeResolverInterface $chain_order_type_resolver,
                              ChainOrderItemTypeResolverInterface $chain_order_item_type_resolver,
                              CartManagerInterface $cart_manager,
                              CartProviderInterface $cart_provider,
                              CurrentStoreInterface $current_store,
                              RouteMatchInterface $route_match,
                              EntityTypeManagerInterface $entity_type_manager,
                              MessengerInterface $messenger,
                              CurrencyFormatter $currency_formatter
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->chainOrderTypeResolver = $chain_order_type_resolver;
    $this->chainOrderItemTypeResolver = $chain_order_item_type_resolver;
    $this->cartManager = $cart_manager;
    $this->cartProvider = $cart_provider;
    $this->currentStore = $current_store->getStore();
    $this->routeMatch = $route_match;
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
    $this->currencyFormatter = $currency_formatter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('commerce_order.chain_order_type_resolver'),
      $container->get('commerce_order.chain_order_item_type_resolver'),
      $container->get('commerce_cart.cart_manager'),
      $container->get('commerce_cart.cart_provider'),
      $container->get('commerce_store.current_store'),
      $container->get('current_route_match'),
      $container->get('entity_type.manager'),
      $container->get('messenger'),
      $container->get('commerce_price.currency_formatter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function clickSortable() {
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();
    $options['combine'] = ['default' => 1];
    $options['allow_decimal'] = ['default' => 0];
    $options['step'] = ['default' => 1];

    $options['ajax'] = ['default' => 0];
    $options['enhance_widget'] = ['default' => 0];

    $options['subtotal'] = ['default' => 1];
    $options['strip_trailing_zeroes'] = ['default' => 1];
    $options['currency_display'] = ['default' => 'symbol'];

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);

    $form['qty_options'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Quantity settings'),
    ];

    $form['qty_options']['combine'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Combine'),
      '#description' => $this->t('Combine order items containing the same product variation.'),
      '#default_value' => $this->options['combine'],
    ];

    $form['qty_options']['allow_decimal'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Allow decimal quantities'),
      '#default_value' => $this->options['allow_decimal'],
    ];

    $form['qty_options']['step'] = [
      '#type' => 'select',
      '#title' => $this->t('Step'),
      '#description' => $this->t('Only quantities that are multiples of the selected step will be allowed.'),
      '#options' => [
        '0.01' => '0.01',
        '0.05' => '0.05',
        '0.1' => '0.1',
        '0.25' => '0.25',
        '0.5' => '0.5',
      ],
      '#default_value' => $this->options['step'],
      '#states' => [
        'visible' => [
          ':input[name="options[qty_options][allow_decimal]"]' => ['checked' => TRUE],
        ],
      ],
    ];


    $form['form_options'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Form options'),
    ];

    $form['form_options']['ajax'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use AJAX'),
      '#default_value' => $this->options['ajax'],
    ];

    $form['form_options']['enhance_widget'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enhance widget'),
      '#description' => $this->t('Transforms the form elements number with +/- buttons around'),
      '#default_value' => $this->options['enhance_widget'],
      '#states' => [
        'visible' => [
          ':input[name="options[form_options][ajax]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $form['form_options']['subtotal'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Show subtotal'),
      '#description' => $this->t('Appends a subtotal info about Variations added to cart'),
      '#default_value' => $this->options['subtotal'],
    ];

    $form['form_options']['currency_display'] = [
      '#type' => 'select',
      '#title' => $this->t('Subtotal currency display'),
      '#options' => [
        'symbol' => $this->t('Symbol (e.g. "$")'),
        'code' => $this->t('Currency code (e.g. "USD")'),
        'none' => $this->t('None'),
      ],
      '#default_value' => $this->options['currency_display'],
      '#states' => [
        'visible' => [
          ':input[name="options[form_options][subtotal]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $form['form_options']['strip_trailing_zeroes'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Strip trailing zeroes after the decimal point.'),
      '#default_value' => $this->options['strip_trailing_zeroes'],
      '#states' => [
        'visible' => [
          ':input[name="options[form_options][subtotal]"]' => ['checked' => TRUE],
        ],
      ],
    ];

  }

  /**
   * @param $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitOptionsForm(&$form, FormStateInterface $form_state) {
    parent::submitOptionsForm($form, $form_state); // TODO: Change the autogenerated stub

    $options = &$form_state->getValue('options');
    $qty_options = isset($options['qty_options']) ? $options['qty_options'] : [];
    foreach ($qty_options as $key => $value) {
      $this->options[$key] = $value;
    }

    $form_options = isset($options['form_options']) ? $options['form_options'] : [];
    foreach ($form_options as $key => $value) {
      $this->options[$key] = $value;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getValue(ResultRow $row, $field = NULL) {
    return '<!--form-item-' . $this->options['id'] . '--' . $row->index . '-->';
  }

  /**
   * Form constructor for the views form.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function viewsForm(array &$form, FormStateInterface $form_state) {

    $current_product = $this->routeMatch->getParameter('commerce_product');
    if ($current_product) {
      // Make sure we do not accidentally cache this form.
      $form['#cache']['max-age'] = 0;

      // The view is empty, abort.
      if (empty($this->view->result)) {
        unset($form['actions']);
        return;
      }

      $form['#attached'] = [
        'library' => ['commerce_order/total_summary'],
      ];

      $cart = NULL;

      $default_variation = $current_product->getDefaultVariation();
      $currency_code = $default_variation->getPrice()->getCurrencyCode();
      $total_subtotal = 0;
      if ($default_variation) {
        $order_type = $this->getOrderType($default_variation);

        /** @var \Drupal\commerce_order\Entity\OrderInterface $cart */
        $cart = $this->cartProvider->getCart($order_type, $this->currentStore);
        $total_items = 0;
        $defaults = [];
        if ($cart) {
          foreach ($cart->getItems() as $orderItem) {
            /** @var \Drupal\commerce_product\Entity\ProductVariation $prod_variation */
            $prod_variation = $orderItem->getPurchasedEntity();
            $prod = $prod_variation->getProduct();
            if ($prod->id() == $current_product->id()) {
              $total_items += $orderItem->getQuantity();
              $total_subtotal += $orderItem->getTotalPrice()->getNumber();

              $purchasable_item = $orderItem->getPurchasedEntity();
              $defaults[$purchasable_item->id()] = $orderItem->getQuantity();
            }
          }
        }

        $form[$this->options['id']]['#tree'] = TRUE;
        foreach ($this->view->result as $row_index => $row) {
          $step = 1;
          $precision = 0;

          if ($this->options['allow_decimal']) {
            $step = floatval($this->options['step']);
          }
          $default_value = isset($defaults[$row->variation_id]) ?
            round($defaults[$row->variation_id], $precision) : 0;

          $form[$this->options['id']][$row_index] = [
            '#type' => 'number',
            '#title' => $this->t('Quantity'),
            '#title_display' => 'invisible',
            '#default_value' => $default_value,
            '#name' => 'variation:'.$row->variation_id,
            '#size' => 4,
            '#min' => 0,
            '#max' => 9999,
            '#step' => $step,
          ];

          if ($this->options['enhance_widget']) {
            $form['#attached']['library'][] = 'commerce_pvt/plus-minus';
            $form[$this->options['id']][$row_index]['#attributes']['readonly'] = TRUE;

            $saved_field = $form[$this->options['id']][$row_index];
            unset($form[$this->options['id']][$row_index]);

            $form[$this->options['id']][$row_index] = [
              '#type' => 'container',
            ];

            $form[$this->options['id']][$row_index]['minus'] = [
              '#type' => 'submit',
              '#value' => '-',
              '#name' => 'variation:'.$row->variation_id.':minus',
              '#ajax' => [
                'callback' => [get_class($this), 'updatePageElements'],
                'progress' => [
                  'type' => 'throbber',
                  'message' => '',
                ],
              ],
              '#alter_variation' => TRUE,
              '#alter_step' => 0 - $step,
              '#disabled' => ($default_value == 0),
            ];

            $form[$this->options['id']][$row_index]['number'] = $saved_field;

            $form[$this->options['id']][$row_index]['plus'] = [
              '#type' => 'submit',
              '#value' => '+',
              '#name' => 'variation:'.$row->variation_id.':plus',
              '#ajax' => [
                'callback' => [get_class($this), 'updatePageElements'],
                'progress' => [
                  'type' => 'throbber',
                  'message' => '',
                ],
              ],
              '#alter_variation' => TRUE,
              '#alter_step' => $step,
            ];

            $form[$this->options['id']][$row_index]['#theme_wrappers'] = ['commerce_pvt_plus_minus'];
          }
        }

        $form['actions']['submit']['#manage_cart'] = TRUE;

        // Replace the form submit button label.
        if ($cart && $total_items > 0) {
          $form['actions']['submit']['#value'] = $this->t('Update variations quantitites');
          $form['actions']['submit']['#update_cart'] = TRUE;
        }
        else {
          $form['actions']['submit']['#value'] = $this->t('Add variations to cart');
        }

        if ($this->options['ajax']) {
          $form['#prefix'] = '<div class="commerce-pvt-ajax-form-wrapper">';
          $form['#suffix'] = '</div>';

          $form['actions']['submit']['#ajax'] = [
            'callback' => [get_class($this), 'updatePageElements'],
            'progress' => [
              'type' => 'throbber',
              'message' => ($cart && $total_items > 0 ? $this->t('Updating cart') : $this->t('Adding to cart')).'...',
            ]
          ];
        }

        if ($this->options['subtotal']) {
          $form['actions']['#weight'] = 200;

          $form['footer'] = [
            '#type' => 'container',
            '#weight' => 100,
          ];

          $options = $this->getFormatterOptions();

          $form['footer']['subtotal'] = [
            '#markup' => new FormattableMarkup('<span class="label">@label</span>: <span class="value">@value</span>', [
              '@label' => $this->t('Subtotal for this product\'s variations'),
              '@value' => $this->currencyFormatter->format($total_subtotal, $currency_code, $options),
            ]),
            '#prefix' => '<div class="order-total-line">',
            '#suffix' => '</div>',
          ];

        }
      }
    }
  }

  /**
   * Submit handler for the views form.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function viewsFormSubmit(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();

    $update_cart = !empty($triggering_element['#update_cart']);

    $current_product = $this->routeMatch->getParameter('commerce_product');
    $default_variation = $current_product->getDefaultVariation();
    if ($default_variation) {
      $order_type = $this->getOrderType($default_variation);

      /** @var \Drupal\commerce_order\Entity\OrderInterface $cart */
      $cart = $this->cartProvider->getCart($order_type, $this->currentStore);

      if (!empty($triggering_element['#manage_cart'])) {

        $save_cart = FALSE;

        /** @var \Drupal\commerce_order\Entity\OrderInterface $cart */
        $quantities = [];
        foreach ($form_state->getUserInput() as $input_key => $value) {
          if (strpos($input_key, 'variation:') !== FALSE) {
            list($null, $variation_id) = explode(":", $input_key);
            $quantities[$variation_id] = floatval($value);
          }
        }
        $total_adding = array_sum($quantities);

        if (!$cart && $total_adding > 0) {
          $cart = $this->cartProvider->createCart($order_type);
          $save_cart = TRUE;
        }

        $variation_storage = $this->entityTypeManager->getStorage('commerce_product_variation');
        foreach ($quantities as $variation_id => $quantity) {
          if (!is_numeric($quantity)) {
            // The input might be invalid if the #required or #min attributes
            // were removed by an alter hook.
            continue;
          }

          /** @var \Drupal\commerce_order\Entity\OrderItemInterface $order_item */
          $order_item = $this->getOrderItemFromCart($cart, $variation_id);
          if ($update_cart) {
            if ($order_item) {
              if ($quantity == 0) {
                $this->cartManager->removeOrderItem($cart, $order_item, $this->options['combine'], FALSE);
                $save_cart = TRUE;
              }
              else {
                $order_item->setQuantity($quantity);
                $this->cartManager->updateOrderItem($cart, $order_item, FALSE);
                $save_cart = TRUE;
              }
            }
            else {
              if ($quantity > 0) {
                /** @var \Drupal\commerce_product\Entity\ProductVariation $variation */
                $variation = $variation_storage->load($variation_id);
                $order_item = $this->cartManager->createOrderItem($variation, $quantity);
                $this->cartManager->addOrderItem($cart, $order_item, $this->options['combine'], FALSE);
                $save_cart = TRUE;
              }
            }
          }
          else {
            if ($quantity > 0) {
              /** @var \Drupal\commerce_product\Entity\ProductVariation $variation */
              $variation = $variation_storage->load($variation_id);
              $order_item = $this->cartManager->createOrderItem($variation, $quantity);
              $this->cartManager->addOrderItem($cart, $order_item, $this->options['combine'], FALSE);
              $save_cart = TRUE;
            }
          }
        }

        if ($save_cart) {
          $cart->save();
          $this->cartProvider->clearCaches();
          if ($this->options['ajax']) {
            $form_state->setRebuild();  // update form after cart emptying.
          }
          else {
            $this->messenger->addMessage($this->t('Your shopping cart has been updated.'));
          }
        }
      }

      elseif (!empty($triggering_element['#alter_variation'])) {
        list($null, $variation_id, $action) = explode(":", $triggering_element['#name']);

        if (!$cart) {
          $cart = $this->cartProvider->createCart($order_type);
        }

        $variation_storage = $this->entityTypeManager->getStorage('commerce_product_variation');
        /** @var \Drupal\commerce_order\Entity\OrderItemInterface $order_item */
        $order_item = $this->getOrderItemFromCart($cart, $variation_id);
        if ($order_item) {
          $quantity = $order_item->getQuantity();
          $new_quantity = $quantity + $triggering_element['#alter_step'];
          if (intval($new_quantity) === 0) {
            $this->cartManager->removeOrderItem($cart, $order_item);
          }
          else {
            $order_item->setQuantity($new_quantity);
            $this->cartManager->updateOrderItem($cart, $order_item);
          }
        }
        else {
          if ($action == 'plus') {
            /** @var \Drupal\commerce_product\Entity\ProductVariation $variation */
            $variation = $variation_storage->load($variation_id);
            $order_item = $this->cartManager->createOrderItem($variation, 1);
            $this->cartManager->addOrderItem($cart, $order_item, $this->options['combine']);
          }
        }

        $this->cartProvider->clearCaches();

        if ($this->options['ajax']) {
          $form_state->setRebuild();  // update form after cart emptying.
        }
      }
    }
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   * @throws \Exception
   */
  public function updatePageElements(array &$form, FormStateInterface $form_state) {
    /** @var RefreshPageElementsHelper $refresher */
    $refresher = \Drupal::service('commerce_pvt.refresh_page_elements_helper');
    return $refresher->updatePageElements($form)->getResponse();
  }

  /**
   * {@inheritdoc}
   */
  public function query() {
    // Do nothing.
  }


  /**
   * @param \Drupal\commerce_order\Entity\OrderInterface $cart
   * @param $variation_id
   *
   * @return \Drupal\commerce_order\Entity\OrderItemInterface|null
   */
  private function getOrderItemFromCart(OrderInterface $cart, $variation_id) {
    if (!$cart) {
      return NULL;
    }
    foreach ($cart->getItems() as $orderItem) {
      if ($orderItem->getPurchasedEntityId() == $variation_id) {
        return $orderItem;
      }
    }
    return NULL;
  }

  /**
   * Gets the Order Type for which we could take action.
   *
   * @param \Drupal\commerce\PurchasableEntityInterface $purcasable_entity
   *
   * @return string
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getOrderType($purcasable_entity) {
    $order_item_type = $this->chainOrderItemTypeResolver->resolve($purcasable_entity);
    $order_item = $this->entityTypeManager->getStorage('commerce_order_item')
      ->create(['type' => $order_item_type]);
    return $this->chainOrderTypeResolver->resolve($order_item);
  }

  private function getFormatterOptions() {
    $options = [
      'currency_display' => $this->options['currency_display'],
    ];
    if ($this->options['strip_trailing_zeroes']) {
      $options['minimum_fraction_digits'] = 0;
    }

    return $options;
  }

}
